import time, sys
from plumbum import cli
from walt.client import myhelp
from walt.client.link import ClientToServerLink, ResponseQueue
from walt.client.tools import confirm
from walt.client.interactive import run_node_cmd, \
                                    run_device_ping, \
                                    NODE_SHELL_MESSAGE
from walt.client.transfer import run_transfer_with_node

POE_REBOOT_DELAY            = 2  # seconds

myhelp.register_topic('node-terminology', """
* 'owning' a node
* ---------------
In WalT terminology, if node <N> is deployed with an image created by user <U>,
we consider that "<U> owns <N>".

Thus, if you just started using WalT, "you do not own any node" until you deploy
an image on one of them (use 'walt node deploy <node(s)> <image>' for this).

* specifying a set of nodes 
* -------------------------
Some commands accept a "set of nodes":
- walt node deploy
- walt node reboot
- walt log show         (see option '--nodes')

In this case you can specify either:
* the keyword 'my-nodes' (this will select the nodes that you own)
* the keyword 'all-nodes'
* a coma separated list of nodes (e.g "rpi1,rpi2" or just "rpi1")
""")

class WalTNode(cli.Application):
    """WalT node management sub-commands"""
    @staticmethod
    def confirm_nodes_not_owned(server, node_set):
        not_owned = server.includes_nodes_not_owned(node_set, warn=True)
        if not_owned == None:
            return False
        if not_owned == True:
            if not confirm():
                return False
        return True

    @staticmethod
    def run_cmd(node_set, several_nodes_allowed, cmdargs, startup_msg=None):
        nodes_ip = None
        with ClientToServerLink() as server:
            if not WalTNode.confirm_nodes_not_owned(server, node_set):
                return
            nodes_ip = server.get_reachable_nodes_ip(node_set)
            if len(nodes_ip) == 0:
                return  # issue already reported
            elif len(nodes_ip) > 1 and not several_nodes_allowed:
                sys.stderr.write(
                    'Error: this command must target 1 node only.\n')
                return
        if nodes_ip:
            for ip in nodes_ip:
                if startup_msg:
                    print startup_msg
                run_node_cmd(ip, cmdargs)

@WalTNode.subcommand("show")
class WalTNodeShow(cli.Application):
    """show WalT nodes"""
    _all = False # default
    def main(self):
        with ClientToServerLink() as server:
            print server.show_nodes(self._all)
    @cli.autoswitch(help='show nodes used by other users too')
    def all(self):
        self._all = True

@WalTNode.subcommand("blink")
class WalTNodeBlink(cli.Application):
    """make a node blink for a given number of seconds"""
    def main(self, node_name, duration=60):
        try:
            seconds = int(duration)
        except:
            sys.stderr.write(
                '<duration> must be an integer (number of seconds).\n')
        else:
            with ClientToServerLink() as server:
                if server.blink(node_name, True):
                    print 'blinking for %ds... ' % seconds
                    try:
                        time.sleep(seconds)
                        print 'done.'
                    except KeyboardInterrupt:
                        print 'Aborted.'
                    finally:
                        server.blink(node_name, False)

@WalTNode.subcommand("reboot")
class WalTNodeReboot(cli.Application):
    """reboot a (set of) node(s)"""
    def main(self, node_set):
        with ClientToServerLink() as server:
            if not WalTNode.confirm_nodes_not_owned(server, node_set):
                return
            if server.poweroff(node_set, warn_unknown_topology=True):
                time.sleep(POE_REBOOT_DELAY)
                server.poweron(node_set, warn_unknown_topology=False)

@WalTNode.subcommand("deploy")
class WalTNodeDeploy(cli.Application):
    """deploy an operating system image on a (set of) node(s)"""
    def main(self, node_set, image_name):
        with ClientToServerLink() as server:
            if server.has_image(image_name):
                if not WalTNode.confirm_nodes_not_owned(server, node_set):
                    return
                if server.poweroff(node_set, warn_unknown_topology=True):
                    server.set_image(node_set, image_name, warn_unknown_topology=False)
                    time.sleep(POE_REBOOT_DELAY)
                    server.poweron(node_set, warn_unknown_topology=False)

@WalTNode.subcommand("ping")
class WalTNodePing(cli.Application):
    """check that a node is reachable on WalT network"""
    def main(self, node_name):
        node_ip = None
        with ClientToServerLink() as server:
            node_ip = server.get_node_ip(node_name)
        if node_ip:
            run_device_ping(node_ip)

@WalTNode.subcommand("shell")
class WalTNodeShell(cli.Application):
    """run an interactive shell connected to the node"""
    def main(self, node_name):
        WalTNode.run_cmd(   node_name, False, ['bash'],
                            startup_msg=NODE_SHELL_MESSAGE)

@WalTNode.subcommand("run")
class WalTNodeRun(cli.Application):
    """run a command on a (set of) node(s)"""
    def main(self, node_set, *cmdargs):
        WalTNode.run_cmd(node_set, True, cmdargs)

@WalTNode.subcommand("cp")
class WalTNodeCp(cli.Application):
    """transfer files/dirs (client machine <-> node)"""
    def main(self, src, dst):
        with ClientToServerLink() as server:
            info = server.validate_node_cp(src, dst)
            if info == None:
                return
            info = { k:v for k,v in info }
            try:
                run_transfer_with_node(**info)
            except (KeyboardInterrupt, EOFError):
                print
                print 'Aborted.'

@WalTNode.subcommand("wait")
class WalTNodeWait(cli.Application):
    """wait for a node (or a set of nodes) to be ready"""
    def main(self, node_set):
        q = ResponseQueue()
        try:
            with ClientToServerLink(True) as server:
                server.wait_for_nodes(q, node_set)
                q.wait()
        except KeyboardInterrupt:
            print 'Aborted.'




import json, re
from os.path import expanduser
from collections import OrderedDict
from walt.common.tools import read_json

SINGLE_ITEM=1
ONE_OR_MORE_ITEMS=2

CONF_ITEMS = OrderedDict([
    ('server', dict(
        desc='ip or hostname of walt server',
        itemtype=str,
        default='localhost',
        user_input=True
    )),
    ('username', dict(
        desc='walt username',
        explain='''\
This name is used to indicate the author of the walt images you create.
If you have an account on the docker hub, you should input the same username,
because walt looks for any walt image stored there.
Otherwise you are free to choose the one you want.''',
        itemtype=str,
        default=None,
        user_input=True
    ))
])

def ask_config(key):
    infos = CONF_ITEMS[key]
    desc = infos['desc']
    default = infos['default']
    itemtype = infos['itemtype']
    print
    print 'Missing configuration item:', desc
    if 'explain' in infos:
        print infos['explain']
    if default:
        msg = 'please enter it now [%s]: ' % str(default)
    else:
        msg = 'please enter it now: '
    while True:
        value = raw_input(msg)
        if value.strip() == '':
            if default:
                value = default
                print 'Selected value:', str(value)
                break
            else:
                continue
        try:
            value = itemtype(value)
            break
        except:
            print "Bad input: expected type is '%s'" % str(itemtype)
    return value

def get_config_file():
    return expanduser('~/walt.conf')

def get_config(config_file):
    modified = False
    conf = read_json(config_file)
    if conf == None:
        conf = OrderedDict()
        modified = True
    for key in CONF_ITEMS:
        if key not in conf:
            if CONF_ITEMS[key]['user_input']:
                conf[key] = ask_config(key)
            else:
                conf[key] = CONF_ITEMS[key]['default']
            modified = True
    if modified:
        save_config(conf, config_file)
        print '\nConfiguration was stored in %s.\n' % config_file
    return conf

def save_config(conf, config_file):
    lines=[ ]
    for k in conf:
        # if not the 1st item add ',' before continuing
        if len(lines) > 0:
            lines = lines[0:-1] + [ lines[-1] + ',' ]
        # add item label and description as 'json comments'
        infos = CONF_ITEMS[k]
        lines.append('')
        lines.append('    # %s' % infos['desc'])
        lines.append('    # %s' % re.sub('.', '-', infos['desc']))
        if 'explain' in infos:
            lines.extend([ '    # ' + line for line in infos['explain'].splitlines() ])
        # get json output for this item only
        # (by creating a temporary dictionary with just this item)
        item_and_value = json.dumps({k:conf[k]}, indent=4).splitlines()[1:-1]
        lines.extend(item_and_value)
    # concatenate all and write the file
    with open(config_file, 'w') as f:
        f.write('{' + '\n'.join(lines) + '\n}\n')

conf_path = get_config_file()
conf = get_config(conf_path)




#!/usr/bin/env python
import rpyc, os, sys, signal, threading
from walt.common.constants import WALT_SERVER_DAEMON_PORT
from walt.client.config import conf
from walt.client.filesystem import Filesystem
from multiprocessing import Queue

class ResponseQueue(object):
    def __init__(self):
        self.q = Queue()
    def exposed_put(self, *args, **kwargs):
        self.q.put(*args, **kwargs)
    def exposed_done(self):
        self.q.put(None)
    def get(self):
        return self.q.get()
    def wait(self):
        self.q.get()

class ExposedStream(object):
    def __init__(self, stream):
        self.stream = stream
    def exposed_fileno(self):
        return self.stream.fileno()
    def exposed_readline(self, size=-1):
        return self.stream.readline(size)
    def exposed_write(self, s):
        self.stream.write(s)
    def exposed_flush(self):
        self.stream.flush()

# most of the functionality is provided at the server,
# of course. 
# but the client also exposes a few objects / features
# in the following class.
class WaltClientService(rpyc.Service):
    def __init__(self, *args, **kwargs):
        rpyc.Service.__init__(self, *args, **kwargs)
        self.exposed_stdin = ExposedStream(sys.stdin)
        self.exposed_stdout = ExposedStream(sys.stdout)
        self.exposed_stderr = ExposedStream(sys.stderr)
        self.exposed_username = conf['username']
        self.exposed_filesystem = Filesystem()

# in some cases we need a background thread that will handle
# RPyC events.
class BgRPyCThread(object):
    def __init__(self, conn):
        self.conn = conn
        self.thread = threading.Thread(target = self._bg_server)
        self.thread.setDaemon(True)
        self.active = True
        self.thread.start()
    def __del__(self):
        if self.active:
            self.stop()
    def _bg_server(self):
        try:
            while self.active:
                self.conn.serve()
        except Exception:
            if self.active:
                raise
    def stop(self):
        assert self.active
        self.active = False
        self.thread.join()
        self.conn = None

class ClientToServerLink:
    def __init__(self, bg_thread_enabled = False):
        self.bg_thread_enabled = bg_thread_enabled

    def __enter__(self):
        self.conn = rpyc.connect(
                conf['server'],
                WALT_SERVER_DAEMON_PORT,
                service = WaltClientService)
        if self.bg_thread_enabled:
            self.bg_thread = BgRPyCThread(self.conn)
        return self.conn.root

    def __exit__(self, type, value, traceback):
        if self.bg_thread_enabled:
            self.bg_thread.stop()
        self.conn.close()




from plumbum import cli
from walt.client.link import ClientToServerLink
from walt.client.interactive import run_sql_prompt

class WalTAdvanced(cli.Application):
    """advanced sub-commands"""
    pass

@WalTAdvanced.subcommand("sql")
class WalTAdvancedSql(cli.Application):
    """Start a remote SQL prompt on the WalT server database"""
    def main(self):
        run_sql_prompt()

@WalTAdvanced.subcommand("fix-image-owner")
class WalTAdvancedFixImageOwner(cli.Application):
    """fix the owner of images"""
    _force = False # default
    def main(self, other_user):
        if not self._force:
            print """\
This will make you own all images of user '%s'. It is intended
for maintenance only (i.e. if user '%s' is no longer working with
walt).
If this is really what you want, run:
walt advanced fix-image-owner --yes-i-know-do-it-please %s
""" % ((other_user,) * 3)
        else:
            with ClientToServerLink() as server:
                server.fix_image_owner(other_user)
    @cli.autoswitch(help='yes, I know, do it!')
    def yes_i_know_do_it_please(self):
        self._force = True




from plumbum import cli
from walt.client.link import ClientToServerLink
from walt.client.interactive import run_device_ping

class WalTDevice(cli.Application):
    """management of WalT platform devices"""
    pass

@WalTDevice.subcommand("tree")
class WalTDeviceTree(cli.Application):
    """print the network structure of the platform"""
    def main(self):
        with ClientToServerLink() as server:
            print server.device_tree()

@WalTDevice.subcommand("show")
class WalTDeviceShow(cli.Application):
    """print details about devices involved in the platform"""
    def main(self):
        with ClientToServerLink() as server:
            print server.device_show()

@WalTDevice.subcommand("rescan")
class WalTDeviceRescan(cli.Application):
    """rescan the network devices involved in the platform"""
    def main(self):
        with ClientToServerLink() as server:
            server.device_rescan()

@WalTDevice.subcommand("rename")
class WalTRenameDevice(cli.Application):
    """rename a device"""
    def main(self, old_name, new_name):
        with ClientToServerLink() as server:
            server.rename(old_name, new_name)

@WalTDevice.subcommand("ping")
class WalTDevicePing(cli.Application):
    """check that a device is reachable on WalT network"""
    def main(self, device_name):
        device_ip = None
        with ClientToServerLink() as server:
            device_ip = server.get_device_ip(device_name)
        if device_ip:
            run_device_ping(device_ip)

MSG_FORGET_DEVICE_WITH_LOGS = """\
This would delete any information about %s, including %s log \
lines.
If this is what you want, run 'walt node forget --force %s'."""

MSG_REACHABLE_CANNOT_FORGET = """\
Failed: %s is reachable on the WalT network (thus supposedly in use).
You can use 'walt device rescan' to update this.
"""
@WalTDevice.subcommand("forget")
class WalTDeviceForget(cli.Application):
    """let the WalT system forget about an obsolete device"""
    _force = False # default
    def main(self, device_name):
        with ClientToServerLink() as server:
            reachable = server.is_device_reachable(device_name)
            if reachable == None:
                return  # issue already reported
            if reachable:
                print MSG_REACHABLE_CANNOT_FORGET % device_name
                return
            if not self._force:
                logs_cnt = server.count_logs(
                        history = (None, None),
                        senders = set([device_name]))
                if logs_cnt > 0:
                    print MSG_FORGET_DEVICE_WITH_LOGS % (
                        device_name, logs_cnt, device_name
                    )
                    return  # give up for now
            # ok, do it
            server.forget(device_name)
            print 'done.'
    @cli.autoswitch(help='do it, even if related data will be lost')
    def force(self):
        self._force = True




import sys
from multiprocessing import Process, Queue
from Queue import Empty

def confirm():
    while True:
        print 'Are you sure? (y/n):',
        res = raw_input()
        if res == 'y':
            return True
        if res == 'n':
            return False

class ProgressMessageThread(Process):
    def __init__(self, message):
        Process.__init__(self)
        self.message = message
        self.queue = Queue()
    def __enter__(self):
        self.start()
    def __exit__(self, type, value, traceback):
        self.queue.put(0)
        self.join()
    def print_status(self, status, termination='\r'):
        sys.stdout.write("%s %s%s" % (self.message, status, termination))
        sys.stdout.flush()
    def run(self):
        idx = 0
        try:
            while True:
                self.print_status("\\|/-"[idx])
                idx = (idx+1) % 4
                try:
                    self.queue.get(block = True, timeout = 0.2)
                except Empty:
                    continue
                # there is input in the queue, we should notify that we
                # are done and stop
                self.print_status("done", '\n')
                break
        except KeyboardInterrupt:
            self.queue.put(0)




#!/usr/bin/env python
"""
WalT (wireless testbed) control tool.
"""
import sys, socket
from plumbum import cli
from walt.client import myhelp
from walt.client.log import WalTLog
from walt.client.advanced import WalTAdvanced
from walt.client.node import WalTNode
from walt.client.device import WalTDevice
from walt.client.image import WalTImage

WALT_VERSION = "0.1"

class WalT(cli.Application):
    """WalT (wireless testbed) control tool."""
    VERSION = WALT_VERSION

    @cli.switch(["-z", "--help-about"], str,
                group = "Meta-switches")
    def help_about(self, topic):
        """Prints help details about a given topic.
           Run 'walt --help-about help' to list them.
        """
        print myhelp.get(topic)
        sys.exit()

WalT.subcommand("advanced", WalTAdvanced)
WalT.subcommand("device", WalTDevice)
WalT.subcommand("image", WalTImage)
WalT.subcommand("log", WalTLog)
WalT.subcommand("node", WalTNode)

def run():
    try:
        WalT.run()
    except socket.error:
        print 'Network connection to WalT server failed.'

if __name__ == "__main__":
    run()




from walt.client.config import conf
from walt.client.tools import ProgressMessageThread
from walt.common.tcp import write_pickle, client_socket, \
                            Requests
from walt.common.constants import WALT_SERVER_TCP_PORT
import os, tarfile


def run_transfer_with_image(session, dst_dir, dst_name, src_dir, src_name,
                            client_operand_index, **kwargs):
    image_fullname, container_name = session.get_parameters()
    if client_operand_index == 0:
        req_id = Requests.REQ_TAR_TO_IMAGE
    else:
        req_id = Requests.REQ_TAR_FROM_IMAGE
    run_transfer(req_id, dst_dir, dst_name, src_dir, src_name,
                    client_operand_index,
                    container_name = container_name,
                    image_fullname = image_fullname)

def run_transfer_with_node(node_ip, dst_dir, dst_name, src_dir, src_name,
                            client_operand_index, **kwargs):
    if client_operand_index == 0:
        req_id = Requests.REQ_TAR_TO_NODE
    else:
        req_id = Requests.REQ_TAR_FROM_NODE
    run_transfer(req_id, dst_dir, dst_name, src_dir, src_name,
                    client_operand_index,
                    node_ip = node_ip)

def run_transfer(req_id, dst_dir, dst_name, src_dir, src_name,
                            client_operand_index, **entity_params):
    params = dict(
        dst_dir = dst_dir,
        dst_name = dst_name,
        src_dir = src_dir,
        src_name = src_name,
        **entity_params
    )
    # connect to server
    s = client_socket(conf['server'], WALT_SERVER_TCP_PORT)
    f = s.makefile('r+', 0)
    # send the request id
    Requests.send_id(f, req_id)
    # wait for the READY message from the server
    f.readline()
    # write the parameters
    write_pickle(params, f)
    # handle client-side archiving / unarchiving
    with ProgressMessageThread('Transfering...'):
        if client_operand_index == 0:
            # client is sending
            with tarfile.open(mode='w|', fileobj=f) as archive:
                archive.add(os.path.join(src_dir, src_name), arcname=dst_name)
        else:
            # client is receiving
            with tarfile.open(mode='r|', fileobj=f) as archive:
                archive.extractall(path=dst_dir)
    f.close()
    s.close()




#!/usr/bin/env python
from plumbum import cli
from walt.client.link import ClientToServerLink, ResponseQueue
from walt.client.tools import confirm
from walt.client.interactive import run_image_shell_prompt
from walt.client.transfer import run_transfer_with_image

class WalTImage(cli.Application):
    """management of WalT-nodes operating system images"""
    pass

@WalTImage.subcommand("search")
class WalTImageSearch(cli.Application):
    """search for remote WalT node OS images"""
    def main(self, keyword=None):
        with ClientToServerLink(True) as server:
            q = ResponseQueue()
            server.search_images(q, keyword)
            print q.get()

@WalTImage.subcommand("clone")
class WalTImageClone(cli.Application):
    """clone a remote image into your working set"""
    _force = False # default
    def main(self, clonable_image_link):
        q = ResponseQueue()
        with ClientToServerLink(True) as server:
            server.clone_image(q, clonable_image_link, self._force)
            q.wait()
    @cli.autoswitch(help='do it, even if it overwrites an existing image.')
    def force(self):
        self._force = True

@WalTImage.subcommand("show")
class WalTImageShow(cli.Application):
    """display your working set of walt images"""
    def main(self):
        with ClientToServerLink() as server:
            print server.show_images()

@WalTImage.subcommand("shell")
class WalTImageShell(cli.Application):
    """modify an image through an interactive shell"""
    def main(self, image_name):
        with ClientToServerLink() as server:
            session = server.create_image_shell_session(
                            image_name)
            if session == None:
                return  # issue already reported
            with session:
                run_image_shell_prompt(session)
                default_new_name = session.get_default_new_name()
                try:
                    while True:
                        new_name = raw_input(\
                            'New image name [%s]: ' % default_new_name)
                        if new_name == '':
                            new_name = default_new_name
                            print 'Selected: %s' % new_name
                        res = session.validate_new_name(new_name)
                        if res == session.NAME_NEEDS_CONFIRM:
                            if confirm():
                                res = session.NAME_OK
                            else:
                                res = session.NAME_NOT_OK
                        if res == session.NAME_OK:
                            break
                        if res == session.NAME_NOT_OK:
                            continue
                    # we left the loop, this means we have a valid name
                    session.select_new_name(new_name)
                except (KeyboardInterrupt, EOFError):
                    print 'Aborted.'

@WalTImage.subcommand("remove")
class WalTImageRemove(cli.Application):
    """remove an image from your working set"""
    def main(self, image_name):
        with ClientToServerLink() as server:
            server.remove_image(image_name)

@WalTImage.subcommand("rename")
class WalTImageRename(cli.Application):
    """rename an image of your working set"""
    def main(self, image_name, new_image_name):
        with ClientToServerLink() as server:
            server.rename_image(image_name, new_image_name)

@WalTImage.subcommand("duplicate")
class WalTImageDuplicate(cli.Application):
    """duplicate an image of your working set"""
    def main(self, image_name, new_image_name):
        with ClientToServerLink() as server:
            server.duplicate_image(image_name, new_image_name)

@WalTImage.subcommand("cp")
class WalTImageCp(cli.Application):
    """transfer files (client machine <-> image)"""
    def main(self, src, dst):
        with ClientToServerLink() as server:
            info = server.validate_image_cp(src, dst)
            if info == None:
                return
            info = { k:v for k,v in info }
            session = server.create_image_shell_session(
                            info['image_tag'])
            if session == None:
                return  # issue already reported
            with session:
                info.update(session = session)
                try:
                    run_transfer_with_image(**info)
                    if info['client_operand_index'] == 0:
                        # client was sending -> image has been modified
                        # save the image under the same name
                        session.select_new_name(session.get_default_new_name())
                except (KeyboardInterrupt, EOFError):
                    print
                    print 'Aborted.'




import sys

def register_topic(topic, message):
    message_per_topic[topic] = message
    message_per_topic['help'] = \
        'Available help topics are:\n%s' % \
        ', '.join(sorted(t for t in message_per_topic if t != 'help'))

def get(topic):
    if topic not in message_per_topic:
        print 'No such help topic.'
        return get('help')
    return message_per_topic[topic]

message_per_topic = {}




#!/usr/bin/env python
import sys, tty, termios, array, fcntl
from walt.client.config import conf
from sys import stdin, stdout
from select import select
from walt.common.constants import WALT_SERVER_TCP_PORT
from walt.common.io import SmartBufferingFileReader, \
                            unbuffered, read_and_copy
from walt.common.tcp import Requests, write_pickle, client_socket
from walt.client import myhelp

SQL_SHELL_MESSAGE = """\
Type \dt for a list of tables.
"""
IMAGE_SHELL_MESSAGE = """\
Notice: this is a limited virtual environment.
Run 'walt --help-about shells' for more info.
"""
NODE_SHELL_MESSAGE = """\
Caution: changes will be lost on next node reboot.
Run 'walt --help-about shells' for more info.
"""

myhelp.register_topic('shells', """
                | walt node shell    | walt image shell
------------------------------------------------------------
persistence     | until the node     | yes
                | reboots (1)        |
------------------------------------------------------------
backend         | the real node      | virtual environment
                |                    | ARM CPU emulation (2)
------------------------------------------------------------
target workflow | testing/debugging  | apply changes
                |                    |
------------------------------------------------------------

(1): Changes are lost on reboot. This ensures that a node booting a
given image will always act the same.

(2): Avoid heavy processing, such as compiling of a large
source code base. In this case, cross-compiling on another machine
and importing the build artefacts in the virtual environment (through
the emulated network) should be the prefered option.
Also, keep in mind that in the virtual environment (docker container)
no services are running (no init process, etc). Actually, the only
process running in this virtual environment when you enter it is the
shell process itself.
""")

class TTYSettings(object):
    def __init__(self):
        self.tty_fd = sys.stdout.fileno()
        # save
        self.saved = termios.tcgetattr(self.tty_fd)
        self.win_size = self.get_win_size()
    def set_raw_no_echo(self):
        # set raw mode
        tty.setraw(self.tty_fd, termios.TCSADRAIN)
        # disable echo
        new = termios.tcgetattr(self.tty_fd)
        new[3] &= ~termios.ECHO
        termios.tcsetattr(self.tty_fd, termios.TCSADRAIN, new)
    def restore(self):
        # return saved conf
        termios.tcsetattr(self.tty_fd, termios.TCSADRAIN, self.saved)
    def get_win_size(self):
        buf = array.array('h', [0, 0, 0, 0])
        fcntl.ioctl(self.tty_fd, termios.TIOCGWINSZ, buf, True)
        return buf

class PromptClient(object):
    def __init__(self, req_id, **params):
        if sys.stdout.isatty() and sys.stdin.isatty():
            self.client_tty = True
            self.tty_settings = TTYSettings()
            # send terminal width and provided parameters
            params.update(
                win_size = self.tty_settings.win_size
            )
        else:
            self.client_tty = False
        # tell server whether we are on a tty
        params.update(client_tty = self.client_tty)
        # connect
        server_host = conf['server']
        s = client_socket(server_host, WALT_SERVER_TCP_PORT)
        # use unbuffered communication
        self.socket_r = s.makefile('r', 0)
        self.socket_w = s.makefile('w', 0)
        # write request id
        Requests.send_id(self.socket_w, req_id)
        # wait for the READY message from the server
        self.socket_r.readline()
        write_pickle(params, self.socket_w)
        # provide read_available() method
        self.stdin_reader = SmartBufferingFileReader(unbuffered(stdin, 'r'))
        self.socket_reader = SmartBufferingFileReader(self.socket_r)

    def run(self):
        # we will wait on 2 file descriptors
        select_args = [ [ stdin, self.socket_r ], [], [ stdin, self.socket_r ] ]

        # this is the main trick when trying to provide a virtual
        # terminal remotely: the client should set its own terminal
        # in 'raw' mode, in order to avoid interpreting any escape
        # sequence (ctrl-r, etc.). These sequences should just be 
        # forwarded up to the terminal running on the server, and 
        # this remote terminal will interpret them.
        # Also, echo-ing what is typed should be disabled at the
        # client side, otherwise it would be echo-ed twice.
        # It is better to consider that the server terminal will
        # handle all outputs, otherwise the synchronization between
        # the output coming from the client (echo) and from the 
        # server (command outputs, prompts) will not be perfect
        # because of network latency. Thus we let the server terminal
        # handle the echo.
        if self.client_tty:
            self.tty_settings.set_raw_no_echo()
        try:
            while True:
                rlist, wlist, elist = select(*select_args)
                if len(elist) > 0:
                    break
                fd = rlist[0].fileno()
                if fd == self.socket_r.fileno():
                    if read_and_copy(self.socket_reader, stdout) == False:
                        break
                else:
                    if read_and_copy(self.stdin_reader, self.socket_w) == False:
                        break
        finally:
            if self.client_tty:
                self.tty_settings.restore()

def run_sql_prompt():
    print SQL_SHELL_MESSAGE
    PromptClient(Requests.REQ_SQL_PROMPT).run()

def run_image_shell_prompt(session):
    print IMAGE_SHELL_MESSAGE
    image_fullname, container_name = session.get_parameters()
    PromptClient(Requests.REQ_DOCKER_PROMPT,
                image_fullname = image_fullname,
                container_name = container_name).run()

def run_node_cmd(node_ip, cmdargs):
    PromptClient(Requests.REQ_NODE_CMD,
                node_ip=node_ip,
                cmdargs=tuple(cmdargs)).run()

def run_device_ping(device_ip):
    PromptClient(Requests.REQ_DEVICE_PING, device_ip=device_ip).run()




import sys, re, datetime, cPickle as pickle
from plumbum import cli
from walt.common.constants import WALT_SERVER_TCP_PORT
from walt.common.tcp import read_pickle, write_pickle, client_socket, \
                            Requests
from walt.client.config import conf
from walt.client.link import ClientToServerLink
from walt.client.tools import confirm
from walt.client import myhelp

DATE_FORMAT_STRING= '%Y-%m-%d %H:%M:%S'
DATE_FORMAT_STRING_HUMAN= '<YYYY>-<MM>-<DD> <hh>:<mm>:<ss>'
DATE_FORMAT_STRING_EXAMPLE= '2015-09-28 15:16:39'

DEFAULT_FORMAT_STRING= \
   '{timestamp:%H:%M:%S.%f} {sender}.{stream} -> {line}'
POSIX_TIMESTAMP_FORMAT = '{timestamp:%s.%f}'

myhelp.register_topic('log-format', """
You may display walt logs in a custom format by using:
$ walt log show --format FORMAT_STRING [other_options...]

FORMAT_STRING should be a python-compatible string formating pattern.
The default is:
%s

A useful example in case of batch processing is to use '%s'
for the timestamp part. It will display it as a POSIX timestamp (i.e. a
float number of seconds since Jan 1, 1970).
""" % (DEFAULT_FORMAT_STRING, POSIX_TIMESTAMP_FORMAT))

myhelp.register_topic('log-history', """
All logs are saved in a database on the walt server. You may retrieve 
them by using:
$ walt log show --history <range> [other_options...]

The <range> may be either 'full', 'none', or have the form
'[<start>]:[<end>]'.
Omitting <start> means the range has no limit in the past.
Omitting <end> means logs up to now should match.
Thus the range ':' is equivalent to using the keyword 'full'.
 
If specified, <start> and <end> boundaries must be either:
- a relative offset to the current time, in the form '-<num><unit>',
  such as '-40s', '-5m', '-1h', '-10d' (resp. seconds, minutes, hours and days).
- the name of a checkpoint (see 'walt --help-about log-checkpoint')
""")

myhelp.register_topic('log-realtime', """
WalT logs may be retrieved in pseudo-realtime by using:
$ walt log show --realtime [other_options...]

In this mode, 'walt log show' will wait for incoming log records and
display them.

Note that you may combine --realtime and --history options, e.g.:
$ walt log show --history -2min: --realtime
This will display the logs from 2 minutes in the past to now and then
continue listening for new incoming logs.
""")

myhelp.register_topic('log-checkpoint', """
A checkpoint is a kind of marker in time. It allows to easily
reference the associated point in time by just giving its name.

Here is a sample workflow:
$ walt log add-checkpoint exp1-start
[... run experience exp1 ...]
$ walt log add-checkpoint exp1-end
$ walt log show --history exp1-start:exp1-end > exp1.log

By using 'walt log add-checkpoint' with option '--date',
you can create a checkpoint referencing a given date. For example:
$ walt log add-checkpoint --date '%s' my-checkpoint
""" % DATE_FORMAT_STRING_EXAMPLE)

SECONDS_PER_UNIT = {'s':1, 'm':60, 'h':3600, 'd':86400}
NUM_LOGS_CONFIRM_TRESHOLD = 1000

def validate_checkpoint_name(name):
    return re.match('^[a-zA-Z0-9]+[a-zA-Z0-9\-]+$', name)

class LogsFlowFromServer(object):
    def __init__(self, walt_server_host):
        s = client_socket(walt_server_host, WALT_SERVER_TCP_PORT)
        self.f_read = s.makefile('r', 0)
        self.f_write = s.makefile('w', 0)
    def read_log_record(self):
        return read_pickle(self.f_read)
    def request_log_dump(self, **kwargs):
        Requests.send_id(self.f_write, Requests.REQ_DUMP_LOGS)
        write_pickle(kwargs, self.f_write)
    def close(self):
        self.f_read.close()
        self.f_write.close()

class WalTLog(cli.Application):
    """management of logs"""
    pass

@WalTLog.subcommand("show")
class WalTLogShow(cli.Application):
    """Dump logs on standard output"""
    format_string = cli.SwitchAttr(
                "--format",
                str,
                argname = 'LOG_FORMAT',
                default = DEFAULT_FORMAT_STRING,
                help= """format used to print logs (see walt --help-about log-format)""")
    realtime = cli.Flag(
                "--realtime",
                default = False,
                help= """enable realtime mode (see walt --help-about log-realtime)""")
    history_range = cli.SwitchAttr(
                "--history",
                str,
                argname = 'HISTORY_RANGE',
                default = 'none',
                help= """history range to be retrieved (see walt --help-about log-history)""")
    set_of_nodes = cli.SwitchAttr(
                "--nodes",
                str,
                argname = 'SET_OF_NODES',
                default = 'my-nodes',
                help= """targeted nodes (see walt --help-about node-terminology)""")
    streams = cli.SwitchAttr(
                "--streams",
                str,
                argname = 'REGEXP',
                default = None,
                help= """selected log streams (as a regular expr.)""")

    def analyse_history_range(self, server, server_time):
        MALFORMED=(False,)
        try:
            self.history_range = self.history_range.lower()
            if self.history_range == 'none':
                return True, None
            elif self.history_range == 'full':
                return True, (None, None)
            parts = self.history_range.split(':')
            if len(parts) != 2:
                return MALFORMED
            history = []
            for part in parts:
                if part == '':
                    history.append(None)
                elif part.startswith('-'):
                    delay = datetime.timedelta(
                                seconds = int(part[1:-1]) * \
                                    SECONDS_PER_UNIT[part[-1]])
                    history.append(pickle.dumps(server_time - delay))
                elif validate_checkpoint_name(part):
                    cptime = server.get_pickled_checkpoint_time(part)
                    if cptime == None:
                        return MALFORMED
                    history.append(cptime)
                else:
                    return MALFORMED
            if history[0] and history[1]:
                if pickle.loads(history[0]) > pickle.loads(history[1]):
                    print 'Issue with the HISTORY_RANGE specified: ' + \
                            'the starting point is newer than the ending point.'
                    return MALFORMED
            return True, tuple(history)
        except Exception as e:
            return MALFORMED

    def main(self):
        if self.realtime == False and self.history_range == 'none':
            print 'You must specify at least 1 of the options --realtime and --history.'
            print "See 'walt --help-about log-realtime' and 'walt --help-about log-history' for more info."
            return
        with ClientToServerLink() as server:
            senders = server.parse_set_of_nodes(self.set_of_nodes)
            if senders == None:
                return
            senders = set(senders) # weakref -> local object
            if self.streams != None:
                # try to validate the regular expression.
                try:
                    re.compile(self.streams)
                except:
                    print 'Invalid regular expression: %s.' % self.streams
                    return
            server_time = pickle.loads(server.get_pickled_time())
            range_analysis = self.analyse_history_range(server, server_time)
            if not range_analysis[0]:
                print '''Invalid HISTORY_RANGE. See 'walt --help-about log-history' for more info.'''
                return
            history_range = range_analysis[1]
            if history_range:
                num_logs = server.count_logs(history = history_range, senders = senders)
                if num_logs > NUM_LOGS_CONFIRM_TRESHOLD:
                    print 'This will display approximately %d log records from history.' % num_logs
                    if not confirm():
                        print 'Aborted.'
                        return
        conn = LogsFlowFromServer(conf['server'])
        conn.request_log_dump(  history = history_range,
                                realtime = self.realtime,
                                senders = senders,
                                streams = self.streams)
        while True:
            try:
                record = conn.read_log_record()
                if record == None:
                    break
                print self.format_string.format(**record)
                sys.stdout.flush()
            except KeyboardInterrupt:
                print
                break
            except Exception as e:
                print 'Could not display the log record.'
                print 'Verify your format string.'
                break

@WalTLog.subcommand("add-checkpoint")
class WalTLogAddCheckpoint(cli.Application):
    """Record a checkpoint (reference point in time)"""
    date = cli.SwitchAttr("--date", str, default=None)

    def main(self, checkpoint_name):
        if self.date:
            try:
                self.date = pickle.dumps(datetime.datetime.strptime(\
                                self.date, DATE_FORMAT_STRING))
            except:
                print 'Could not parse the date specified.'
                print 'Expected format is: %s' % DATE_FORMAT_STRING_HUMAN
                print 'Example: %s' % DATE_FORMAT_STRING_EXAMPLE
                return
        if not validate_checkpoint_name(checkpoint_name):
            sys.stderr.write("""\
Invalid checkpoint name:
* Only alnum and dash(-) characters are allowed.
* dash(-) is not allowed as the 1st character.
""")
            return
        with ClientToServerLink() as server:
            server.add_checkpoint(checkpoint_name, self.date)

@WalTLog.subcommand("remove-checkpoint")
class WalTLogRemoveCheckpoint(cli.Application):
    """Remove a checkpoint"""
    def main(self, checkpoint_name):
        with ClientToServerLink() as server:
            server.remove_checkpoint(checkpoint_name)

@WalTLog.subcommand("list-checkpoints")
class WalTLogListCheckpoints(cli.Application):
    """List checkpoints"""
    def main(self):
        with ClientToServerLink() as server:
            server.list_checkpoints()



from setuptools import setup, find_packages
import sys

if not sys.version_info[0] == 2:
    sys.exit("Sorry, Python 3 is not supported (yet)")

setup(
    name = "walt-client",
    version = "0.9-2",
    packages = find_packages(),
    install_requires = ['rpyc>=3.3','plumbum>=1.4.2',
                        'walt-common'],

    # metadata for upload to PyPI
    author = "Etienne Duble",
    author_email = "etienne.duble@imag.fr",
    description = "WalT (Wireless Testbed) control tool.",
    license = "LGPL",
    keywords = "WalT wireless testbed",
    url = "http://walt.forge.imag.fr/",

    namespace_packages = ['walt'],
    entry_points = {
        'console_scripts': [
            'walt = walt.client.client:run'
        ]
    },
)




import os

class Filesystem:
    def __init__(self):
        self.exposed_get_file_type = self.get_file_type
    def get_file_type(self, path):
        if not os.path.exists(path):
            return None
        if os.path.isfile(path):
            return 'f'
        if os.path.isdir(path):
            return 'd'
        return 'o'
